/*
                             Joystick Calibrator
 */

#ifndef JC_H
#define JC_H


#include <limits.h>
#include <sys/types.h>

#include <gtk/gtk.h>
#include <jsw.h>


#ifndef gbool
# define gbool	guchar
#endif


#ifndef GTK_EVENTS_FLUSH
# define GTK_EVENTS_FLUSH	\
{ \
 while(gtk_events_pending()) \
  gtk_main_iteration(); \
}
#endif	/* GTK_EVENTS_FLUSH */



/*
 *	Program name and version:
 */
#define PROG_NAME	"Joystick Calibrator"
#define PROG_VERSION	"1.4.2"

#define PROG_VERSION_MAJOR	1
#define PROG_VERSION_MINOR	4
#define PROG_VERSION_RELEASE	2


/*
 *	Program usage message:
 */
#define PROG_USAGE_MESG	"\
Usage: jscalibrator [options] [GUI_options]\n\
\n\
    [options] can be any of the following:\n\
\n\
        -d <device>            Specifies the initial joystick device to\n\
                               open at startup (default /dev/js0).\n\
        -f <file>              Specifies the joystick calibration file\n\
                               to open and save to (default ~/.joystick).\n\
        --help                 Prints (this) help screen and exits.\n\
        --version              Prints version information and exits.\n\
\n\
    [GUI_options] can be any options standard to your GUI, consult\n\
    your GUI's manual for available options.\n\
\n"


/*
 *	Copyright message:
 */
#define PROG_COPYRIGHT  "\
Copyright (C) 1997-2001 WolfPack Entertainment.\n\
This program is protected by international copyright laws and treaties,\n\
distribution and/or modification of this software in violation of the\n\
GNU Public License is strictly prohibited. Violators will be prosicuted\n\
to the fullest extent of the law."


/*
 *	Default joystick device name:
 */
#define JC_DEF_JS_DEVICE	"/dev/js0"

/*
 *	Default joystick calibration file name:
 *
 *	This is looked for in the user's home directory.
 */
#define JC_DEF_JS_CALIB		".joystick"


/*
 *	Default toplevel size for JC window:
 */
#define JC_DEF_WIDTH	600
#define JC_DEF_HEIGHT	350

/*
 *	Default size of joystick button:
 */
#define JC_DEF_JS_BUTTON_WIDTH	80
#define JC_DEF_JS_BUTTON_HEIGHT	25

/*
 *	Default size of calibrate toggle button:
 */
#define JC_DEF_CALIBRATE_TOGGLE_WIDTH	80
#define JC_DEF_CALIBRATE_TOGGLE_HEIGHT	25

/*
 *	Default status bar sizes:
 */
#define JC_DEF_STATUS_BAR_HEIGHT	25



/*
 *	Command codes:
 */
#define JC_CMD_NONE			0

#define JC_CMD_OPEN_CALIB		10
#define JC_CMD_OPEN_DEVICE		11
#define JC_CMD_SAVE_CALIB               12
#define JC_CMD_SAVEAS_CALIB             13
#define JC_CMD_EXIT                     14

#define JC_CMD_SPECIFICATIONS		20
#define JC_CMD_REFRESH                  21
#define JC_CMD_CLOSE_DEVICE		22

#define JC_CMD_LAYOUT_REPRESENTATIVE	30
#define JC_CMD_LAYOUT_LOGICAL		31

#define JC_CMD_HELP_CALIBRATE		40
#define JC_CMD_HELP_ABOUT		41



/*
 *	Cursors:
 */
typedef struct {

	GdkCursor	*processing;	/* Hourglass or watch. */

} jc_cursor_struct;
extern jc_cursor_struct jc_cursor;


/*
 *	Representative joystick axises layout widgets structure:
 */
typedef struct {

	GtkWidget	*toplevel,	/* Hbox inside notebook. */
			*instructions_label,
			*axis_da,	/* First two axises drawing area. */
			*throttle_da,
			*rotate_da,
			*hat_da,
			*calib_toggle;

	GdkPixmap	*axis_buf,	/* First two axises. */
			*throttle_buf,
			*rotate_buf,
			*hat_buf;

	GdkColor	c_axis_bg,	/* Background for axises. */
			c_axis_fg,	/* Foreground for axises. */
			c_axis_grid,
			c_axis_hat,
			c_throttle_fg;	/* Throttle axis fg. */

} layout_rep_struct;


/*
 *	Logical joystick axises layout widgets structure:
 */
typedef struct {

	GtkWidget	*toplevel,	/* Parent vbox. */
			*axis_num_label,
			*position_da,
			*calib_toggle,
			*nz_spin,
			*is_hat_check,
			*flipped_check;

	GtkObject	*nz_adj;

	GdkPixmap	*position_buf;

	GdkColor	c_axis_bg,	/* Background for axises. */
			c_axis_fg,	/* Foreground for axises. */
			c_axis_grid;

} layout_logical_axis_struct;

typedef struct {

	GtkWidget       *toplevel;	/* Hbox inside notebook. */

	layout_logical_axis_struct **axis;
	gint total_axises;

} layout_logical_struct;


/*
 *	Joystick button widgets structure:
 */
typedef struct {

	GtkWidget *toplevel;		/* Parent vbox. */

	/* Button label widgets, corresponds with number of buttons. */
	GtkWidget **button;
	gint total_buttons;

} buttons_list_struct;


/*
 *	Status bar structure:
 */
typedef struct {

        GtkWidget	*toplevel,	/* Parent vbox. */
			*js_icon_fixed,	/* Dipicts js being init/shutdown. */
			*js_icon_pixmap,
			*mesg_label;

} status_bar_struct;


/*
 *	Joystick calibrator main window:
 */
typedef struct {

	GdkColormap	*colormap;	/* From toplevel. */

	GtkWidget	*toplevel,
			*menu_bar,
			*js_device_combo,
			*signal_level_da,	/* JS signal meter drawing area. */
			*axises_notebook;	/* Notebook holds axis layout widgets. */

	GdkPixmap	*signal_level_pixmap;

	GdkColor	c_signal_level_bg,
			c_signal_level_fg;

	GdkGC *gc;

	/* Layout state, this matches the axises_notebook page number. */
#define JC_LAYOUT_REPRESENTATIVE	0
#define JC_LAYOUT_LOGICAL		1
	gint layout_state;

	/* Axis layout substructures. */
	/* Logical axis layout widgets and structure, the
	 * layout_logical_vp is a scrolled window that is the parent
	 * for layout_logical_parent. The layout_logical_parent is
	 * the parent for each axis widget group's parent widget.
	 */
	GtkWidget		*layout_logical_vp,	/* Viewport. */
				*layout_logical_parent;	/* Parent vbox. */
	layout_logical_struct	layout_logical;

	/* Representative axis layout widgets and structure, the
	 * layout_rep_parent is a GtkVBox parent for 
	 * layout_rep_parent_client. The layout_rep_parent_client is
	 * created and destroyed whenever the axises change.
	 */
        GtkWidget		*layout_rep_parent,
				*layout_rep_parent_client;	/* Parent vbox. */
	layout_rep_struct	layout_rep;


	/* Buttons list. */
	GtkWidget		*buttons_list_vp,	/* Viewport. */
				*buttons_list_parent;	/* Parent vbox. */
	buttons_list_struct buttons_list;

	/* Status bar. */
	status_bar_struct status_bar;


	/* Calibration file. */
	gchar *calib_file;
	gboolean tried_load_device_on_init;

	/* Modifications marker. */
	gboolean has_changes;

	/* Joystick data. */
	js_data_struct jsd;


	/* Joystick signal load level history. */
#define JC_SIGNAL_HISTORY_MAX   30
	u_int16_t signal_history[JC_SIGNAL_HISTORY_MAX];

} jc_struct;
extern jc_struct js_calib;


/*
 *	File browser:
 */
extern GtkWidget *filebrowser;


/* In jccalib.c */
extern void JCCalibrateRepresentative(jc_struct *jc);
extern void JCDoCalibrate(jc_struct *jc);
extern void JCDoSetAxisTolorance(jc_struct *jc, gint axis_num);


/* In jccb.c */
extern void JCSignalHandler(gint s);
extern gint JCExitCB(
        GtkWidget *widget, GdkEvent *event, gpointer data
);
extern void JCMenuCB(GtkWidget *widget, gpointer data);
extern void JCSwitchPageCB(
	GtkNotebook *notebook, GtkNotebookPage *page, guint page_num,
        gpointer data
);
extern void JCJSDeviceEntryCB(GtkWidget *widget, gpointer data);
extern void JCRefreshCB(GtkWidget *widget, gpointer data);
extern void JCCalibToggleCB(GtkWidget *widget, gpointer data);
extern void JCCalibToggleHighlightCB(GtkWidget *widget, gpointer data);
extern void JCCalibToggleUnhighlightCB(GtkWidget *widget, gpointer data);
extern void JCNZSpinCB(GtkWidget *widget, gpointer data);
extern void JCCalibIsHatCheckCB(GtkWidget *widget, gpointer data);
extern void JCCalibFlipCheckCB(GtkWidget *widget, gpointer data);
extern gint JCAxisExposeCB(
        GtkWidget *widget, GdkEventExpose *event, gpointer data
);
extern gint JCTimeoutCB(gpointer data);

extern void JCDoPrintSpecifications(jc_struct *jc);

extern gint JCDoOpenCalibration(
        jc_struct *jc, const gchar *path
);
extern void JCJSOpenDeviceCB(gpointer data);
extern gint JCDoOpenJoystick(
	jc_struct *jc, const gchar *path
);
extern void JCDoCloseJoystick(jc_struct *jc);

extern void JCLockProcessing(jc_struct *jc);
extern void JCUnlockProcessing(jc_struct *jc);


/* In jcdraw.c */
extern void JCDrawAxises(jc_struct *jc);
extern void JCDrawButtons(jc_struct *jc);
extern void JCDrawSignalLevel(jc_struct *jc);


/* In jcfb.c */
extern void JCFBOpenCalibCB(GtkWidget *w, GtkFileSelection *fs);
extern void JCFBSaveCalibCB(GtkWidget *w, GtkFileSelection *fs);
extern void JCFBCancelCB(GtkWidget *w, GtkFileSelection *fs);
extern void JCFBDestroyCB(GtkWidget *widget, GtkFileSelection *fs);

extern gint JCFBCreate(GtkWidget **fb, gint argc, gchar **argv);
extern void JCFBMapByCommand(GtkWidget **fb, gint command, jc_struct *jc);
extern void JCFBDestroy(GtkWidget **fb);


/* In jcfile.c */
extern int JCDoSaveCalibration(jc_struct *jc, const char *path);


/* In jcmanage.c */
extern gint JCCreate(jc_struct *jc, gint argc, gchar **argv);
extern void JCMap(jc_struct *jc);
extern void JCUnmap(jc_struct *jc);
extern void JCDestroy(jc_struct *jc);


/* In jcwidgets.c */
extern gint JCCreateRepresentativeLayoutWidgets(
        jc_struct *jc, layout_rep_struct *lr
);
extern void JCDestroyRepresentativeLayoutWidgets(
        jc_struct *jc, layout_rep_struct *lr
);
extern gint JCCreateLogicalLayoutWidgets(
        jc_struct *jc, layout_logical_struct *ll,
        GtkWidget *parent       /* Hbox inside notebook. */
);
extern void JCUpdateLogicalLayoutWidgets(
        jc_struct *jc, layout_logical_struct *ll
);
extern void JCDestroyLogicalLayoutWidgets(
        jc_struct *jc, layout_logical_struct *ll
);
extern gint JCCreateButtonsList(jc_struct *jc);
extern void JCDestroyButtonsList(jc_struct *jc);

extern void JCUpateDeviceComboList(jc_struct *jc);
extern gint JCCreateWidgets(jc_struct *jc, gint argc, gchar **argv);
extern void JCDestroyWidgets(jc_struct *jc);


/* In main.c */
extern gint JCInit(gint argc, gchar **argv);
extern void JCShutdown(void);


/* In statusbar.c */
extern void StatusBarSetJSIcon(
        status_bar_struct *sb, u_int8_t **icon_data
);
extern void StatusBarSetJSState(status_bar_struct *sb, gint state);
extern void StatusBarSetMesg(status_bar_struct *sb, const gchar *mesg);
extern gint StatusBarInit(status_bar_struct *sb, GtkWidget *parent);
extern void StatusBarDestroy(status_bar_struct *sb);


#endif	/* JC_H */
