#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "jc.h"


void JCCalibrateRepresentative(jc_struct *jc);
void JCCalibrateLogical(jc_struct *jc);
void JCDoCalibrate(jc_struct *jc);

void JCDoSetAxisTolorance(jc_struct *jc, gint axis_num);


/*
 *	Calibrate axises in representative layout.
 */
void JCCalibrateRepresentative(jc_struct *jc)
{
	int i;
	js_axis_struct *axis_ptr;
	js_data_struct *jsd;
	GtkWidget *w;
	layout_rep_struct *lr;


	lr = &jc->layout_rep;
	if(lr == NULL)
	    return;

	w = lr->calib_toggle;
	if(w == NULL)
	    return;

	if(!(GTK_TOGGLE_BUTTON(w)->active))
	    return;

	jsd = &jc->jsd;

	for(i = 0; i < jsd->total_axises; i++)
	{
	    axis_ptr = jsd->axis[i];
	    if(axis_ptr == NULL)
		continue;

	    if(axis_ptr->min > axis_ptr->cur)
		axis_ptr->min = axis_ptr->cur;

	    if(axis_ptr->max < axis_ptr->cur)
                axis_ptr->max = axis_ptr->cur;
	}

	return;
}

/*
 *	Calibrate axises in logical layout.
 */
void JCCalibrateLogical(jc_struct *jc)
{
	int i;
	js_axis_struct *axis_ptr;
	js_data_struct *jsd;
	GtkWidget *w;
	layout_logical_struct *ll;
	layout_logical_axis_struct *ll_axis_ptr;


        ll = &jc->layout_logical;
        if(ll == NULL)
            return;

        jsd = &jc->jsd;

        for(i = 0; i < ll->total_axises; i++)
        {
	    ll_axis_ptr = ll->axis[i];
	    if(ll_axis_ptr == NULL)
		continue;

	    w = ll_axis_ptr->calib_toggle;
	    if(w == NULL)
		continue;

	    if(!GTK_TOGGLE_BUTTON(w)->active)
		continue;

	    if(JSIsAxisAllocated(jsd, i))
		axis_ptr = jsd->axis[i];
	    else
		continue;

            if(axis_ptr->min > axis_ptr->cur)
                axis_ptr->min = axis_ptr->cur;

            if(axis_ptr->max < axis_ptr->cur)
                axis_ptr->max = axis_ptr->cur;
	}

	return;
}


/*
 *	Calibrates axis values, should be called onec per cycle.
 */
void JCDoCalibrate(jc_struct *jc)
{
	if(jc == NULL)
	    return;

	switch(jc->layout_state)
	{
	  case JC_LAYOUT_REPRESENTATIVE:
	    JCCalibrateRepresentative(jc);
	    break;

	  case JC_LAYOUT_LOGICAL:
	    JCCalibrateLogical(jc);
	    break;
	}

	return;
}


/*
 *	Procedure to set axis tolorance, may cause process to
 *	`block' for a moment.
 */
void JCDoSetAxisTolorance(jc_struct *jc, gint axis_num)
{
	gint i, status;
	gint total_loops = 15;	/* 1.5 seconds. */
	js_data_struct *jsd;
	js_axis_struct *axis_ptr;
	gint x, e1, e2;


	if(jc == NULL)
	    return;

	jsd = &jc->jsd;
	if(!JSIsInit(jsd))
	    return;

	if(JSIsAxisAllocated(jsd, axis_num))
	    axis_ptr = jsd->axis[axis_num];
	else
	    return;

	JCLockProcessing(jc);

	/* Begin calculating error for tolorance. */
	for(i = 0, x = axis_ptr->cur, e1 = 0;
            i < total_loops;
            i++
	)
	{
	    usleep(100000);

	    /* Get new axis value positions. */
	    status = JSUpdate(jsd);
	    if(status == JSNoEvent)
		continue;

	    /* Calculate differance between last and current axis value. */
	    e2 = x - axis_ptr->cur;
	    if(e2 < 0)
		e2 *= -1;	/* Must be absolute value. */

	    /* New error greater than old? */
	    if(e2 > e1)
		e1 = e2;

	    /* Record last axis value. */
	    x = axis_ptr->cur;
	}

	axis_ptr->tolorance = e1;
/*
g_print("Axis %i error %i\n", axis_num, axis_ptr->tolorance);
 */

	JCUnlockProcessing(jc);

	return;
}
