#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <gtk/gtk.h>

#include "../include/string.h"

#include "cdialog.h"
#include "jc.h"

#include "helpmesgs.h"


void JCSignalHandler(gint s);

gint JCExitCB(GtkWidget *widget, GdkEvent *event, gpointer data);

void JCMenuCB(GtkWidget *widget, gpointer data);
void JCSwitchPageCB(
        GtkNotebook *notebook, GtkNotebookPage *page, guint page_num,
        gpointer data
);
void JCJSDeviceEntryCB(GtkWidget *widget, gpointer data);
void JCRefreshCB(GtkWidget *widget, gpointer data);
void JCJSOpenDeviceCB(gpointer data);
void JCCalibToggleCB(GtkWidget *widget, gpointer data);
void JCCalibToggleHighlightCB(GtkWidget *widget, gpointer data);
void JCCalibToggleUnhighlightCB(GtkWidget *widget, gpointer data);
void JCCalibIsHatCheckCB(GtkWidget *widget, gpointer data);
void JCCalibFlipCheckCB(GtkWidget *widget, gpointer data);
gint JCAxisExposeCB(
	GtkWidget *widget, GdkEventExpose *event, gpointer data
);

gint JCTimeoutCB(gpointer data);

void JCDoPrintSpecifications(jc_struct *jc);

gint JCDoOpenCalibration(
	jc_struct *jc, const gchar *path
);
void JCJSOpenDeviceCB(gpointer data);
gint JCDoOpenJoystick(
	jc_struct *jc, const gchar *path
);
void JCDoCloseJoystick(jc_struct *jc);

void JCLockProcessing(jc_struct *jc);
void JCUnlockProcessing(jc_struct *jc);


#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define MAX(a,b)	(((a) > (b)) ? (a) : (b))


/*
 *	Signal handler.
 */
void JCSignalHandler(gint s)
{
	static int segfault_count = 0;

	switch(s)
	{
          case SIGINT:
          case SIGTERM:
          case SIGKILL:
            /* Must reset has changes before calling JCExitCB(). */
            js_calib.has_changes = FALSE;
            JCExitCB(NULL, NULL, &js_calib);
            break;

          case SIGSEGV:
            segfault_count++;
            fprintf(
                stderr,
                "%s triggered a segmentation fault (%i times)\n",
                PROG_NAME,
                segfault_count
            );
            if(segfault_count >= 3)
            {
                fprintf(stderr,
                    "%s attempting immediate process exit().\n",
                    PROG_NAME
                );
                exit(1);
            }
            break;
	}

	return;
}

/*
 *	Exit callback handler.
 */
gint JCExitCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	int exit_status = 1;
	jc_struct *jc = (jc_struct *)data;
	if(jc != NULL)
	{
	    if(CDialogIsQuery())
	    {
		exit_status = 0;
	    }

	    /* Do comfermation if has changes. */
	    if(jc->has_changes)
	    {
		gint status;

		CDialogSetTransientFor(jc->toplevel);
		status = CDialogGetResponse(
"Discard calibration changes?",
"Current calibration values have not been saved,\n\
are you sure you want to discard the changes?",
"The modifications to the calibration values for\n\
the current joystick device have not been saved,\n\
if you say no then you will be given a chance to\n\
save those changes, otherwise those changes will\n\
be discarded.",
                    CDIALOG_ICON_WARNING,
                    CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO,
                    CDIALOG_BTNFLAG_NO
                );
		CDialogSetTransientFor(NULL);
                switch(status)
                {
                  case CDIALOG_RESPONSE_NO:
	          case CDIALOG_RESPONSE_CANCEL:
		  case CDIALOG_RESPONSE_NOT_AVAILABLE:
		    return(TRUE);
		    break;
		}
	    }
	}

	if(exit_status)
	{
	    /* Deallocate all resources. */
	    JCShutdown();

	    /* Shutdown toolkit. */
	    gtk_main_quit();
	}

	return(TRUE);
}


/*
 *	Menu item activate callback.
 */
void JCMenuCB(GtkWidget *widget, gpointer data)
{
	jc_struct *jc;
	gint command;


	jc = &js_calib;
	command = (gint)data;

	switch(command)
	{
	  /* Do nothing. */
	  case JC_CMD_NONE:
	    break;

	  /* Open calibration file. */
	  case JC_CMD_OPEN_CALIB:
	    /* Map file browser for open calib file operation. */
	    JCFBMapByCommand(&filebrowser, command, jc);
	    break;

	  /* Open joystick device. */
	  case JC_CMD_OPEN_DEVICE:
            if(jc->has_changes)
	    {
                gint status;

		CDialogSetTransientFor(jc->toplevel);
		status = CDialogGetResponse(
"Discard calibration changes?",
"Current calibration values have not been saved,\n\
are you sure you want to discard the changes?",
"The modifications to the calibration values for\n\
the current joystick device have not been saved,\n\
if you say no then you will be given a chance to\n\
save those changes, otherwise those changes will\n\
be discarded.",
                    CDIALOG_ICON_WARNING,
                    CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO,
                    CDIALOG_BTNFLAG_NO
                );
		CDialogSetTransientFor(NULL);
                switch(status)
                {
                  case CDIALOG_RESPONSE_NO:
                  case CDIALOG_RESPONSE_CANCEL:
		  case CDIALOG_RESPONSE_NOT_AVAILABLE:
                    return;
		    break;
		}
	    }
	    JCJSOpenDeviceCB(jc);
	    break;

	  /* Close joystick device. */
	  case JC_CMD_CLOSE_DEVICE:
	    if(jc->has_changes)
	    {
                gint status;

		CDialogSetTransientFor(jc->toplevel);
		status = CDialogGetResponse(
"Discard calibration changes?",
"Current calibration values have not been saved,\n\
are you sure you want to discard the changes?",
"The modifications to the calibration values for\n\
the current joystick device have not been saved,\n\
if you say no then you will be given a chance to\n\
save those changes, otherwise those changes will\n\
be discarded.",
                    CDIALOG_ICON_WARNING,
                    CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO,
                    CDIALOG_BTNFLAG_NO
                );
		CDialogSetTransientFor(NULL);
                switch(status)
                {
                  case CDIALOG_RESPONSE_NO:
                  case CDIALOG_RESPONSE_CANCEL:
                  case CDIALOG_RESPONSE_NOT_AVAILABLE:
                    return;
                    break;
                }
	    }
	    JCDoCloseJoystick(jc);
	    break;

	  /* Print specifications. */
	  case JC_CMD_SPECIFICATIONS:
	    JCDoPrintSpecifications(jc);
	    break;

	  /* Save calibration file. */
	  case JC_CMD_SAVE_CALIB:
	    /* Save current calibration file. */
	    if(jc->calib_file == NULL)
	    {
		JCFBMapByCommand(&filebrowser, JC_CMD_SAVEAS_CALIB, jc);
	    }
	    else
	    {
                gtk_file_selection_set_filename(
                    GTK_FILE_SELECTION(filebrowser),
                    jc->calib_file
		);
		JCFBSaveCalibCB(
		    filebrowser,
		    GTK_FILE_SELECTION(filebrowser)
		);
	    }
	    break;

	  /* Save as alternate calibration file. */
	  case JC_CMD_SAVEAS_CALIB:
	    /* Map file browser for save as calib file operation. */
            JCFBMapByCommand(&filebrowser, command, jc);
	    break;

	  /* Exit the program. */
	  case JC_CMD_EXIT:
            if(jc->has_changes)
	    {
                gint status;

		CDialogSetTransientFor(jc->toplevel);
		status = CDialogGetResponse(
"Discard calibration changes?",
"Current calibration values have not been saved,\n\
are you sure you want to discard the changes?",
"The modifications to the calibration values for\n\
the current joystick device have not been saved,\n\
if you say no then you will be given a chance to\n\
save those changes, otherwise those changes will\n\
be discarded.",
                    CDIALOG_ICON_WARNING,
                    CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO,
                    CDIALOG_BTNFLAG_NO
                );
		CDialogSetTransientFor(NULL);
                switch(status)
                {
                  case CDIALOG_RESPONSE_NO:
                  case CDIALOG_RESPONSE_CANCEL:
                  case CDIALOG_RESPONSE_NOT_AVAILABLE:
                    return;
                    break;
                }
	    }
	    /* Need to reset has changes to false. */
            jc->has_changes = FALSE;
	    JCExitCB(NULL, NULL, jc);
	    break;

	  /* Change joystick axis layout display to representative. */
	  case JC_CMD_LAYOUT_REPRESENTATIVE:
	    gtk_notebook_set_page(
		GTK_NOTEBOOK(jc->axises_notebook),
		JC_LAYOUT_REPRESENTATIVE
	    );
	    break;

          /* Change joystick axis layout display to representative. */
	  case JC_CMD_LAYOUT_LOGICAL:
            gtk_notebook_set_page(
                GTK_NOTEBOOK(jc->axises_notebook),
                JC_LAYOUT_LOGICAL
            );
            break;

	  /* Refresh. */
	  case JC_CMD_REFRESH:
	    JCRefreshCB(jc->toplevel, jc);
	    break;

	  /* Help: Calibration. */
          case JC_CMD_HELP_CALIBRATE:
	    CDialogSetTransientFor(jc->toplevel);
	    CDialogGetResponse(
		"How to caliberate",
		HELP_MESG_CALIBRATE,
		"Instructions on how to caliberate.",
		CDIALOG_ICON_INFO,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
            break;

	  /* Help: About. */
	  case JC_CMD_HELP_ABOUT:
	    if(1)
	    {
		gchar text[1024];

		sprintf(
		    text,
		    "%s\n\nVersion %s\n\n%s",
		    PROG_NAME, PROG_VERSION, PROG_COPYRIGHT
		);
		CDialogSetTransientFor(jc->toplevel);
                CDialogGetResponse(
                    PROG_NAME,
		    text,
		    "About this program.",
                    CDIALOG_ICON_INFO,
                    CDIALOG_BTNFLAG_OK,
                    CDIALOG_BTNFLAG_OK
                );
		CDialogSetTransientFor(NULL);
	    }
	    break;

	  /* Unsupported command. */
	  default:

	    break;
	}
}

/*
 *	Notebook switch page callback.
 */
void JCSwitchPageCB(
        GtkNotebook *notebook, GtkNotebookPage *page, guint page_num,
        gpointer data
)
{
	static gbool reenterant = FALSE;
	jc_struct *jc = (jc_struct *)data;
        if((jc == NULL) || (notebook == NULL))
            return;

	if(reenterant)
	    return;
	else
	    reenterant = TRUE;

	/* Axises notebook? */
	if((void *)notebook == (void *)jc->axises_notebook)
	{
	    /* Set layout state (which always matches page number). */
	    jc->layout_state = page_num;
	}

	reenterant = FALSE;

	return;
}

/*
 *	Joystick device entry activate callback. This handles activate
 *	signals recieved from the jc combo's entry widget.
 */
void JCJSDeviceEntryCB(GtkWidget *widget, gpointer data)
{
	jc_struct *jc = (jc_struct *)data;
	if(jc == NULL)
	    return;

	if(jc->has_changes)
	{
	    gint status;

	    CDialogSetTransientFor(jc->toplevel);
	    status = CDialogGetResponse(
"Discard calibration changes?",
"Current calibration values have not been saved,\n\
are you sure you want to discard the changes?",
"The modifications to the calibration values for\n\
the current joystick device have not been saved,\n\
if you say no then you will be given a chance to\n\
save those changes, otherwise those changes will\n\
be discarded.",
                CDIALOG_ICON_WARNING,
                CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO,
                CDIALOG_BTNFLAG_NO
            );
	    CDialogSetTransientFor(NULL);
            switch(status)
            {
              case CDIALOG_RESPONSE_NO:
              case CDIALOG_RESPONSE_CANCEL:
              case CDIALOG_RESPONSE_NOT_AVAILABLE:
                return;
                break;
            }
	}

	/* Open newly specified joystick device on jc combo's entry
	 * widget then update resources and values. Currently opened
	 * joystick will be closed first.
	 */
	JCJSOpenDeviceCB(jc);

	return;
}

/*
 *	Refresh procedure callback.
 */
void JCRefreshCB(GtkWidget *widget, gpointer data)
{
	static gbool reenterant = FALSE;
        jc_struct *jc = (jc_struct *)data;
        if(jc == NULL)
            return;

	if(reenterant)
	    return;
	else
	    reenterant = TRUE;

/* Need to add more things to refresh here. */

	/* Update listing of joysticks in jc's js device combo list. */
	JCUpateDeviceComboList(jc);

	/* Refresh values on logical axises layout widgets. */
	JCUpdateLogicalLayoutWidgets(
	    jc, &jc->layout_logical
	);



	reenterant = FALSE;

	return;
}

/*
 *	Calibrate toggle callback, these start/end calibration sequences
 *	when a calibrate toggle button from any of the axises layout
 *	widgets structures is toggled.
 *
 *	If the layout is set to representative then and calibration
 *	toggle is switched back off, then each axis will have its assumed
 *	orientation, style, flags, etc values set up.
 */
void JCCalibToggleCB(GtkWidget *widget, gpointer data)
{
        static gbool reenterant = FALSE;
	layout_rep_struct *lr;
	layout_logical_struct *ll;
	jc_struct *jc = (jc_struct *)data;
        if((widget == NULL) || (jc == NULL))
            return;

        if(reenterant)
            return;
        else
            reenterant = TRUE;

	/* Handle by currently selected axises layout page. */
	switch(jc->layout_state)
	{
          /* ******************************************************* */
	  case JC_LAYOUT_REPRESENTATIVE:
	    lr = &jc->layout_rep;
	    if((widget != lr->calib_toggle) &&
               (widget != NULL)
            )
	        break;

	    if(GTK_TOGGLE_BUTTON(widget)->active)
	    {
		/* Begin calibration. */
		int i;
		js_axis_struct *axis_ptr;
		js_data_struct *jsd = &jc->jsd;


		/* Reset all axis limit values. */
                for(i = 0; i < jsd->total_axises; i++)
                {
                    axis_ptr = jsd->axis[i];
                    if(axis_ptr == NULL)
                        continue;

                    axis_ptr->min = axis_ptr->cur;
                    axis_ptr->max = axis_ptr->cur;
                    axis_ptr->cen = axis_ptr->cur;

                    axis_ptr->flags |= JSAxisFlagTolorance;
                    axis_ptr->tolorance = 0;
		}
                /* Sync values for tolorance in error correction specified
                 * in the current axis structures to the low-level joystick
                 * driver.
                 */
		JSResetAllAxisTolorance(jsd);


		/* Update has changes marker. */
		if(!jc->has_changes)
		    jc->has_changes = TRUE;

                StatusBarSetMesg(
		    &jc->status_bar,
 "Calibrating: Move all axises to extremas and then center"
		);
	    }
	    else
	    {
		/* Calibrating done, process new calibrated values in
		 * representative layout.
		 *
		 * Note that representative layout is very imprecise, so
		 * we will be setting many assumed values and some that
		 * the user may not like. If user dosen't like it then
		 * he or she should use logical layout calibration.
		 */
		gint i, len1, len2, len;
		js_axis_struct *axis_ptr;
		js_data_struct *jsd;
		char text[256];


                lr = &jc->layout_rep;

                jsd = &jc->jsd;

                StatusBarSetMesg(
                    &jc->status_bar,
 "Processing new values (do not touch the device)..."
                );

                /* Itterate through each axis. */
                for(i = 0; i < jsd->total_axises; i++)
                {
                    axis_ptr = jsd->axis[i];
                    if(axis_ptr == NULL)
                        continue;

		    /* Set new center. */             
                    axis_ptr->cen = axis_ptr->cur;

		    /* Update status message for this axis. */
                    sprintf(text,
   "Axis %i: Processing calibration values...",
                        i
                    );
                    StatusBarSetMesg(&jc->status_bar, text);

		    GTK_EVENTS_FLUSH

		    /* Adjust null zone. */
		    len1 = MAX(axis_ptr->max - axis_ptr->cen, 0);
		    len2 = MAX(axis_ptr->cen - axis_ptr->min, 0);
		    len = MAX(len1, len2);
		    axis_ptr->nz = len * 0.25;	/* 25% of len is nz. */

		    /* Begin setting other values based on the axis
		     * number.
		     *
		     *	0	x
		     *  1	y
		     *	2	throttle
		     */
		    switch(i)
		    {
		      case 0:	/* Axis #0. */
			/* Nothing significant about this axis. */
			break;

		      case 1:	/* Axis #1. */
			/* This is usually the Y or up/down axis, and
			 * most of the time it needs to be flipped.
			 */
			axis_ptr->flags |= JSAxisFlagFlipped;
			break;

		      case 2:	/* Axis #2. */
			/* This is sometimes the `twist' or `rotate',
			 * while other times it is the throttle. To best
			 * guess we check the total number of axises.
			 */
			switch(jsd->total_axises)
			{
			  case 5:	/* Have x, y, thr, hat. */
			  case 3:	/* Have x, y, thr. */
			    /* This is probably the throttle. */
			    axis_ptr->nz = 0;	/* No nz for throttle. */
			    axis_ptr->flags |= JSAxisFlagFlipped;
			    break;

			  case 6:	/* Have x, y, rot, thr, hat. */
			  case 4:	/* Have x, y, rot, thr. */
			  default:
			    /* All else assume rotate, nothing to do for
			     * rotate.
			     */
			    break;
			}
			break;

		      case 3:	/* Axis #3. */
                        /* This is sometimes the hat or throttle. To best
                         * guess we check the total number of axises.
                         */
                        switch(jsd->total_axises)
                        {
                          case 5:	/* Have x, y, thr, hat. */
			    /* This is probably the hat x. */
			    axis_ptr->flags |= JSAxisFlagIsHat;
			    break;

                          case 6:       /* Have x, y, rot, thr, hat. */
                          case 4:       /* Have x, y, rot, thr. */
			  default:	/* More than 6 axises total. */
                            /* This is probably the throttle. */
                            axis_ptr->nz = 0;   /* No nz for throttle. */
                            axis_ptr->flags |= JSAxisFlagFlipped;
                            break;
                        }
			break;

		      case 4:	/* Axis #4. */
			/* Must be hat x. */
			axis_ptr->flags |= JSAxisFlagIsHat;
			break;

		      case 5:	/* Axis #5. */
			/* Must be hat y. */
			axis_ptr->flags |= JSAxisFlagFlipped;	/* Also flipped. */
                        axis_ptr->flags |= JSAxisFlagIsHat;
			break;

		      default:	/* Beyond the 6th axis. */
			/* Not understood so make no assumptions and
			 * do not set any special values to it.
			 */
			break;
		    }

		    /* Calculate tolorance for error correction. */
		    JCDoSetAxisTolorance(jc, i);

		    GTK_EVENTS_FLUSH
                }
		/* Sync values for tolorance in error correction specified
		 * in the current axis structures to the low-level joystick
		 * driver.
		 */
		JSResetAllAxisTolorance(jsd);

		/* Need to update logical axis widget values based on
		 * newly set axis flags and options set here.
		 */
		JCUpdateLogicalLayoutWidgets(jc, &jc->layout_logical);

                StatusBarSetMesg(
                    &jc->status_bar,
                    "New calibration values set"
                );
	    }
	    break;

	  /* ******************************************************* */
	  case JC_LAYOUT_LOGICAL:
            ll = &jc->layout_logical;
            if((ll != NULL) && (widget != NULL))
	    {
		gint i;
		layout_logical_axis_struct *ll_axis_ptr;
		js_data_struct *jsd;
		js_axis_struct *axis_ptr;
		char text[256];


                jsd = &jc->jsd;

		/* Go through axises, search for which axis
		 * is to be toggled.
		 */
		for(i = 0; i < ll->total_axises; i++)
		{
		    ll_axis_ptr = ll->axis[i];
		    if(ll_axis_ptr == NULL)
			continue;

		    if(ll_axis_ptr->calib_toggle != widget)
			continue;

		    if(JSIsAxisAllocated(jsd, i))
			axis_ptr = jsd->axis[i];
		    else
			continue;

                    /* Update has changes marker. */
                    if(!jc->has_changes)
                        jc->has_changes = TRUE;

                    if(GTK_TOGGLE_BUTTON(widget)->active)
                    {
			/* Reset values and enter calibration for
			 * this axis.
			 */
			axis_ptr->cen = axis_ptr->cur;
                        axis_ptr->min = axis_ptr->cur;
                        axis_ptr->max = axis_ptr->cur;

			axis_ptr->flags |= JSAxisFlagTolorance;
			axis_ptr->tolorance = 0;

                        /* Sync values for tolorance in error correction
			 * specified in the current axis structures to the
			 * low-level joystick driver.
			 */
			JSResetAllAxisTolorance(jsd);

                        sprintf(text,
   "Calibrating axis %i: Move axis to extremas and then center",
			    i
			);
		    }
		    else
		    {
			/* Finished calibration of this axis. */

                        StatusBarSetMesg(
                            &jc->status_bar,
 "Processing new values (do not touch the device)..." 
                        );
			GTK_EVENTS_FLUSH

                        /* Set new center. */
                        axis_ptr->cen = axis_ptr->cur;

                        /* Calculate tolorance for error correction. */
                        JCDoSetAxisTolorance(jc, i);

			GTK_EVENTS_FLUSH

                        /* Sync values for tolorance in error correction
                         * specified in the current axis structures to the
                         * low-level joystick driver.
                         */
			JSResetAllAxisTolorance(jsd);

                        sprintf(text,
   "Axis %i: Calibration values set",
                            i
                        );
		    }
		    StatusBarSetMesg(&jc->status_bar, text);
		}
	    }
	    break;
	}

	reenterant = FALSE;

	return;
}

/*
 *      Callback to the logical layout is hat check.
 */
void JCNZSpinCB(GtkWidget *widget, gpointer data)
{
        gint i;
        js_axis_struct *axis_ptr;
        layout_logical_struct *ll;
        layout_logical_axis_struct *ll_axis_ptr;
        jc_struct *jc = (jc_struct *)data;
        if(jc == NULL)
            return;

	/* Not on the logical axis layout page? */
        if(jc->layout_state != JC_LAYOUT_LOGICAL)
            return;

        ll = &jc->layout_logical;
        if(ll == NULL)
            return;

	/* Go through each axis. */
        for(i = 0; i < ll->total_axises; i++)
        {
            ll_axis_ptr = ll->axis[i];
            if(ll_axis_ptr == NULL)
                continue;

            if((gpointer)widget != (gpointer)ll_axis_ptr->nz_adj)
                continue;

            if(JSIsAxisAllocated(&jc->jsd, i))
                axis_ptr = jc->jsd.axis[i];
            else
                continue;

	    axis_ptr->nz = gtk_spin_button_get_value_as_int(
		GTK_SPIN_BUTTON(ll_axis_ptr->nz_spin)
	    );
        }

	return;
}

/*
 *	Toggle button highlight callback, this just changes the
 *	status message.
 */
void JCCalibToggleHighlightCB(GtkWidget *widget, gpointer data)
{
	gint i;
	layout_rep_struct *lr;
	layout_logical_struct *ll;
	layout_logical_axis_struct *ll_axis_ptr;
        jc_struct *jc = (jc_struct *)data;
        if((widget == NULL) || (jc == NULL))
            return;


	/* Check which button was highlighted. */

	/* Check layout representative first. */
	lr = &jc->layout_rep;
	if(widget == lr->calib_toggle)
	{
            StatusBarSetMesg(
		&jc->status_bar,
		"Click here to calibrate joystick"
	    );
	}

	/* Check layout logical. */
	ll = &jc->layout_logical;
	for(i = 0; i < ll->total_axises; i++)
	{
            ll_axis_ptr = ll->axis[i];
            if(ll_axis_ptr == NULL)
                continue;

	    if(widget == ll_axis_ptr->calib_toggle)
	    {
		char text[256];

		sprintf(text, "Click here to calibrate axis #%i", i);
		StatusBarSetMesg(&jc->status_bar, text);
	    }
	}

	return;
}

/*
 *      Toggle button highlight callback, this just clears the
 *      status message.
 */
void JCCalibToggleUnhighlightCB(GtkWidget *widget, gpointer data)
{
        gint i;
        layout_rep_struct *lr;
        layout_logical_struct *ll;
        layout_logical_axis_struct *ll_axis_ptr;
        jc_struct *jc = (jc_struct *)data;
        if((widget == NULL) || (jc == NULL)) 
            return;
            
        
        /* Check which button was unhighlighted. */

        /* Check layout representative first. */   
        lr = &jc->layout_rep;
        if(widget == lr->calib_toggle)
            StatusBarSetMesg(&jc->status_bar, NULL);

        /* Check layout logical. */
        ll = &jc->layout_logical;
        for(i = 0; i < ll->total_axises; i++)
        {
            ll_axis_ptr = ll->axis[i];
            if(ll_axis_ptr == NULL)
                continue;

            if(widget == ll_axis_ptr->calib_toggle)
                StatusBarSetMesg(&jc->status_bar, NULL);
        }

        return;
}

/*
 *	Callback to the logical layout is hat check.
 */
void JCCalibIsHatCheckCB(GtkWidget *widget, gpointer data)
{
	gint i;
	js_axis_struct *axis_ptr;
	layout_logical_struct *ll;
	layout_logical_axis_struct *ll_axis_ptr;
	jc_struct *jc = (jc_struct *)data;
	if(jc == NULL)
	    return;

	/* Not on the logical axises layout page? */
	if(jc->layout_state != JC_LAYOUT_LOGICAL)
	    return;

	ll = &jc->layout_logical;

	/* Go through each axis. */
	for(i = 0; i < ll->total_axises; i++)
	{
	    ll_axis_ptr = ll->axis[i];
	    if(ll_axis_ptr == NULL)
		continue;

	    if(widget != ll_axis_ptr->is_hat_check)
		continue;

            if(JSIsAxisAllocated(&jc->jsd, i))
                axis_ptr = jc->jsd.axis[i];
            else
                continue;

	    if(GTK_TOGGLE_BUTTON(widget)->active)
		axis_ptr->flags |= JSAxisFlagIsHat;
	    else
		axis_ptr->flags &= ~JSAxisFlagIsHat;
	}

	return;
}

/*
 *	Callback to the logical layout flipped check.
 */
void JCCalibFlipCheckCB(GtkWidget *widget, gpointer data)
{
        gint i;
        js_axis_struct *axis_ptr;
        layout_logical_struct *ll;
        layout_logical_axis_struct *ll_axis_ptr;
        jc_struct *jc = (jc_struct *)data;
        if(jc == NULL)
            return;

        /* Not on the logical axises layout page? */
        if(jc->layout_state != JC_LAYOUT_LOGICAL)
            return;
 
        ll = &jc->layout_logical;

	/* Go through each axis. */                
        for(i = 0; i < ll->total_axises; i++)
        {       
            ll_axis_ptr = ll->axis[i];
            if(ll_axis_ptr == NULL)
                continue;
  
            if(widget != ll_axis_ptr->flipped_check)
                continue;

            if(JSIsAxisAllocated(&jc->jsd, i))
                axis_ptr = jc->jsd.axis[i];
            else
                continue;

            if(GTK_TOGGLE_BUTTON(widget)->active)
                axis_ptr->flags |= JSAxisFlagFlipped;
            else
                axis_ptr->flags &= ~JSAxisFlagFlipped;
	}

	return;
}

/*
 *	Axis expose event callback.
 */
gint JCAxisExposeCB(
	GtkWidget *widget, GdkEventExpose *event,
	gpointer data
)
{
        jc_struct *jc;

	jc = (jc_struct *)data;
	if(jc == NULL)
	    return(FALSE);

	JCDrawAxises(jc);

	return(FALSE);
}

/*
 *	Timeout callback, this function is called once per loop.
 */
gint JCTimeoutCB(gpointer data)
{
        gint i, n, status;
        u_int16_t *history;
	jc_struct *jc;
	js_data_struct *jsd;



	jc = (jc_struct *)&js_calib;
	jsd = &jc->jsd;

	/* ********************************************************** */
	/* Open joystick device for the first time as needed. */
	if(!jc->tried_load_device_on_init)
	{
	    GtkCombo *combo = (GtkCombo *)jc->js_device_combo;

	    if((combo != NULL) && (jsd->fd < 0))
	    {
		JCDoOpenJoystick(
		    jc,
		    gtk_entry_get_text(GTK_ENTRY(combo->entry))
		);
	    }

	    jc->tried_load_device_on_init = TRUE;
	}

	/* ********************************************************** */
        /* `Shift' joystick signal load history. */
        history = jc->signal_history;
        for(i = 0; i < (JC_SIGNAL_HISTORY_MAX - 1); i++)
            history[i] = history[i + 1];

	/* Get address of last history position. */
        history = &jc->signal_history[JC_SIGNAL_HISTORY_MAX - 1];


	/* Get new joystick values/event. */
	if(jsd->fd > -1)
	    status = JSUpdate(jsd);
	else
	    status = JSNoEvent;

	/* Check if a joystick event was recieved. */
	if(status == JSGotEvent)
	{
	    /* Perform calibrations if needed. */
	    JCDoCalibrate(jc);

	    /* Redraw axises and buttons. */
	    JCDrawAxises(jc);
	    JCDrawButtons(jc);

	    /* Update signal load history. */
	    n = (*history) + 1000;
	    if(n > (gint)((u_int16_t)-1))
		n = (gint)((u_int16_t)-1);
            (*history) = n;
	}
	else
	{
            /* Update signal load history. */
            n = (*history) - 1000;
            if(n < 0)
                n = 0;
	    (*history) = n;
	}

	/* ********************************************************** */
	/* Redraw signal load level. */
	JCDrawSignalLevel(jc);


	return(TRUE);
}

/*
 *	Prints specifications of currently opened joystick.
 */
void JCDoPrintSpecifications(jc_struct *jc)
{
	if(jc == NULL)
	    return;


	if(JSIsInit(&jc->jsd))
	{
	    int version_major, version_minor, version_minor_minor;
	    gint len = 2048;
            gchar *buf;
	    js_data_struct *jsd;


	    jsd = &jc->jsd;

	    if(jsd->name != NULL)
		len += strlen(jsd->name);

	    /* Get parsed version. */
	    JSDriverQueryVersion(
		jsd, &version_major, &version_minor, &version_minor_minor
	    );

	    /* Allocate buffer to print specifications on. */
	    buf = (gchar *)g_malloc(len * sizeof(gchar));
	    if(buf == NULL)
		return;

	    sprintf(
		buf,
"Descriptive Name:\n\
    \"%s\"\n\
\n\
Driver Version: %i.%i.%i\n\
\n\
Device File: %s\n\
Calibration File: %s\n\
\n\
Total Axises: %i\n\
Total Buttons: %i\n\
\n\
Has Modifications: %s",
		((jsd->name == NULL) ? "(null)" : jsd->name),
		version_major, version_minor, version_minor_minor,
		((jsd->device_name == NULL) ? "(null)" : jsd->device_name),
		((jsd->calibration_file == NULL) ? "(null)" : jsd->calibration_file),
		jsd->total_axises,
		jsd->total_buttons,
		((jc->has_changes) ? "Yes" : "No")
	    );
	    CDialogSetTransientFor(jc->toplevel);
            CDialogGetResponse(
                "Specifications",
		buf,
		"Specifications for current opened joystick.",
                CDIALOG_ICON_INFO,
                CDIALOG_BTNFLAG_OK,
                CDIALOG_BTNFLAG_OK 
            );
	    CDialogSetTransientFor(NULL);

	    g_free(buf);	/* Free buffer used to print specifications on. */
	}
	else
	{
	    CDialogSetTransientFor(jc->toplevel);
            CDialogGetResponse(
"Cannot get specifications",
"Joystick not initialized, cannot get specifications.",
"Specifications could be obtained because the joystick\n\
is not initialized or opened. Verify that the joystick\n\
device is properly initlaized and opened.",
                CDIALOG_ICON_ERROR,
                CDIALOG_BTNFLAG_OK,
                CDIALOG_BTNFLAG_OK 
            );
	    CDialogSetTransientFor(NULL);
        }
}

/*
 *	Procedure to open a new calibration file.
 */
gint JCDoOpenCalibration(
	jc_struct *jc, const gchar *path
)
{
	struct stat stat_buf;


        if((jc == NULL) ||
           (path == NULL)
	)
            return(-1);


	/* Make sure path exists. */
	if(stat(path, &stat_buf))
	{
	    gchar text[PATH_MAX + NAME_MAX + 256];
	    gchar tmp_path[PATH_MAX + NAME_MAX];


	    strncpy(tmp_path, path, PATH_MAX + NAME_MAX);
	    tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

	    sprintf(text,
"No such file:\n\n\
    %s",
		tmp_path
	    );
	    CDialogSetTransientFor(jc->toplevel);
            CDialogGetResponse(
"Cannot open calibration file",
text,
"The calibration could not be found at the specified location\n\
please verify that it exists and the location is correct.",
                CDIALOG_ICON_ERROR,
                CDIALOG_BTNFLAG_OK,
                CDIALOG_BTNFLAG_OK 
            );
	    CDialogSetTransientFor(NULL);

            sprintf(
                text,
                "Could not open calibration file"
            );
            StatusBarSetMesg(&jc->status_bar, text);

	    return(-1);
	}


        /* Close currently opened joystick if any. */
        JCDoCloseJoystick(jc);


	/* Set new calibration file name. */
	if(path != jc->calib_file)
	{
	    free(jc->calib_file);
	    jc->calib_file = StringCopyAlloc(path);
	}

	/* Set startup initialized joystick to FALSE so that the joystick
	 * gets initialized in the timeout callback.
	 */
	jc->tried_load_device_on_init = FALSE;


	return(0);
}

/*
 *	Open joystick callback, fetches joystick device path and passes
 *	it to JCDoOpenJoystick(). 
 */
void JCJSOpenDeviceCB(gpointer data)
{
	gchar *strptr;
	GtkCombo *combo;
	jc_struct *jc = (jc_struct *)data;
	if(jc == NULL)
	    return;

	combo = (GtkCombo *)jc->js_device_combo;
	if(combo != NULL)
	    strptr = gtk_entry_get_text(GTK_ENTRY(combo->entry));
	else
	    strptr = NULL;

	if(strptr != NULL)
	{
	    gchar path[PATH_MAX + NAME_MAX - 1];

            /* Copy text from js device entry, JCDoOpenJoystick() needs
             * it done this way.
             */
            strncpy(path, strptr, PATH_MAX + NAME_MAX);
            path[PATH_MAX + NAME_MAX - 1] = '\0';

            /* Open the joystick. */
            JCDoOpenJoystick(jc, path);
	}

	return;  
} 


/*
 *	Procedure to open a joystick. Any joystick currently opened on
 *	the specified jc structure will be first closed with a call to
 *	JCDoCloseJoystick(). The has_changes member in the jc structure
 *	will be reset to FALSE.
 *
 *	The joystick specified by path will then  be attempted to be
 *	opened, any errors encountered will be printed on the message
 *	dialog. If opening was successful then all resources and messages
 *	will be updated to reflect the newly opened joystick.
 *
 *	Returns non-zero on error.
 */
gint JCDoOpenJoystick(jc_struct *jc, const gchar *path)
{
	gchar *strptr;
	GtkCombo *combo;
	gint status;


	if((jc == NULL) ||
	   (path == NULL)
	)
	    return(-1);

	/* Reset has changes on jc. */
	if(jc->has_changes)
	    jc->has_changes = FALSE;

	/* Close currently opened joystick if any, this will also
	 * destroy any axis layout widgets and the buttons in the
	 * buttons list. Plus any other related resources that need
	 * to be destroyed.
	 */
	JCDoCloseJoystick(jc);

	GTK_EVENTS_FLUSH

	/* Open the joystick, the value of status will contain the
	 * success or error code.
	 */
        status = JSInit(
            &jc->jsd,
            path,
            jc->calib_file,
	    JSFlagNonBlocking
        );
	/* Cannot open joystick? */
	if(status != JSSuccess)
        {
	    /* Print message indicating the error code. */
	    gchar text[PATH_MAX + NAME_MAX + 1024];

            switch(status)
            {
              case JSBadValue:
                sprintf(text,
"In invalid value was encountered while attempting to initialize the\n\
joystick. Please check that all values needed to initialize the\n\
joystick are specified and specified within their ranges (ie file paths\n\
and joystick value limits)."
                );
                break;

              case JSNoAccess:
                sprintf(text,
"Could not access one or more specified paths. Check the specified\n\
paths, make sure that they exist and that their permissions allow\n\
you to access them. Also check to make sure the joystick in question\n\
is actually connected, turned on (as needed), and the driver or module\n\
is loaded (type `modprobe <driver_name>')."
                );
                break;

              case JSNoBuffers:
                sprintf(text,
"Insufficient memory available to initialize joystick."
                );  
                break;

              default:  /* JSError */
                sprintf(text,
"Unknown error code `%i' recieved while initializing joystick:\n\n\
    %s",
                    status, path
                );
                break;
            }
	    CDialogSetTransientFor(jc->toplevel);
            CDialogGetResponse(
"Cannot open joystick device",
text,
"The joystick device could not be opened due to the\n\
specified error.",
                CDIALOG_ICON_ERROR,
                CDIALOG_BTNFLAG_OK,
                CDIALOG_BTNFLAG_OK 
            );
	    CDialogSetTransientFor(NULL);

	    sprintf(
		text,
		"Cound not open joystick device `%s'",
		path
	    );
	    StatusBarSetMesg(&jc->status_bar, text);

	    return(-1);
	}

	/* Joystick opened successfully, begin updating values. */

	/* Update the joystick device combo's entry value because the
	 * given path may differ from the one currently specified on
	 * the combo's entry widget.
	 */
	combo = (GtkCombo *)jc->js_device_combo;
	if(combo != NULL)
	    strptr = gtk_entry_get_text(GTK_ENTRY(combo->entry));
	else
	    strptr = NULL;

	if(path != strptr)
	{
	    gtk_entry_set_text(
		GTK_ENTRY(combo->entry),
		path
	    );
	}


	/* Destroy axises and buttons widgets and related resources. */
	JCDestroyRepresentativeLayoutWidgets(
	    jc, &jc->layout_rep
	);
	JCDestroyLogicalLayoutWidgets(
	    jc, &jc->layout_logical
	);
	JCDestroyButtonsList(jc);

	/* Create axises and button widgets which should not reflect
	 * the values of the newly opened joystick specified on the
	 * jsd structure on the jc.
	 */
	JCCreateRepresentativeLayoutWidgets(
	    jc, &jc->layout_rep
	);
        JCCreateLogicalLayoutWidgets(
            jc, &jc->layout_logical,
            jc->layout_logical_parent
        );
	JCCreateButtonsList(jc);


	/* Update jc toplevel window title to contain the name of
	 * joystick device fetched from the newly opened jsd structure.
	 */
	if(jc->jsd.name != NULL)
	{
	    gchar title[256 + 80];
	    gchar js_name[80];

	    strncpy(js_name, jc->jsd.name, 80);
	    js_name[80 - 1] = '\0';

	    sprintf(title, "%s: %s", PROG_NAME, js_name);

	    gtk_window_set_title(
		GTK_WINDOW(jc->toplevel),
		title
	    );
	}

	/* Redraw axis and buttons layout widgets. */
	JCDrawAxises(jc);
	JCDrawButtons(jc);

	/* Update information on status bar. */
	StatusBarSetJSState(&jc->status_bar, 1);
        StatusBarSetMesg(&jc->status_bar, "Joystick initialized");

	return(0);
}

/*
 *	Procedure to close a joystick. This will close the (if opened)
 *	joystick device specified on the jc's jsd structure, any
 *	widgets on the axis layouts will be destroyed as well.
 *
 *	All values will be updated to reflect the closed joystick.
 */
void JCDoCloseJoystick(jc_struct *jc)
{
	GtkWidget *w;

	if(jc == NULL)
	    return;

        /* Reset has changes marker. */
        if(jc->has_changes)
            jc->has_changes = FALSE;

        /* Destroy axises and buttons widgets and related resources
	 * on the axises layout structures and buttons list.
	 */
        JCDestroyRepresentativeLayoutWidgets(
            jc, &jc->layout_rep
        );
        JCDestroyLogicalLayoutWidgets(
            jc, &jc->layout_logical
        );
        JCDestroyButtonsList(jc);


	/* Close the joystick device. */
        JSClose(&jc->jsd);

	/* Update jc's toplevel window title. */
	w = jc->toplevel;
	if(w != NULL)
	{
	    gtk_window_set_title(
	        GTK_WINDOW(jc->toplevel),
		PROG_NAME
	    );
	}

        /* Redraw axis and buttons layout. */
        JCDrawAxises(jc);
        JCDrawButtons(jc);

        /* Update information on status bar. */
        StatusBarSetJSState(&jc->status_bar, 0);
        StatusBarSetMesg(&jc->status_bar, "Joystick shutdown");

	return;
}


/*
 *	Locks out input for processing, this basically displays
 *	the processing cursor.
 *
 *	JCUnlockProcessing() must be called to allow input again.
 */
void JCLockProcessing(jc_struct *jc)
{
	if(jc == NULL)
	    return;

	if(jc->toplevel == NULL)
	    return;

	if(jc_cursor.processing != NULL)
	    gdk_window_set_cursor(
		jc->toplevel->window,
		jc_cursor.processing
	    );

	gdk_flush();

	return;
}

/*
 *	Unlocks processing lockup, basically reversing what
 *	JCLockProcessing() did.
 */
void JCUnlockProcessing(jc_struct *jc)
{
        if(jc == NULL)
            return;

        if(jc->toplevel == NULL)
            return;

        gdk_window_set_cursor(
            jc->toplevel->window,
            NULL
        );

        gdk_flush();

	return;
}
