#include <stdio.h>
#include <gtk/gtk.h>
#include "jc.h"


static js_axis_struct *JCGetAxis(js_data_struct *jsd, gint i);

void JCDrawAxisesRepresentative(jc_struct *jc);
void JCDrawAxisesLogical(jc_struct *jc);
void JCDrawAxises(jc_struct *jc);
void JCDrawButtons(jc_struct *jc);

void JCDrawSignalLevel(jc_struct *jc);


#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define MAX(a,b)        (((a) > (b)) ? (a) : (b))


/*
 *	Returns the axis structure i from the jsd structure.
 *
 *	Can return NULL if axis number i is invalid or on error.
 */
static js_axis_struct *JCGetAxis(js_data_struct *jsd, gint i)
{
	if(jsd == NULL)
	    return(NULL);
	else if((i < 0) || (i >= jsd->total_axises))
	    return(NULL);
	else
	    return(jsd->axis[i]);
}


/*
 *	Redraws the axises layout on the given jc
 *	for the representative layout.  The given inputs are
 *	assumed valid and the layout is assumed to be representative.
 */
void JCDrawAxisesRepresentative(jc_struct *jc)
{ 
	gint x, y, width, height;
	GtkWidget *w;
	GdkPixmap *pixmap;
	GdkGC *gc;
	layout_rep_struct *lr;
	js_axis_struct *axis_ptr1, *axis_ptr2;
	gfloat	axis0_value, axis1_value, axis2_value, axis3_value,
		axis4_value, axis5_value;
	int total_axises;


	lr = &jc->layout_rep;

	/* Not initialized? */
	if(lr->toplevel == NULL)
	    return;

	gc = jc->gc;
	if(gc == NULL)
	    return;


	/* Get first four axis values (set value 0 if they don't
	 * exist.
	 */
	total_axises = jc->jsd.total_axises;
        axis0_value = JSGetAxisCoeff(&jc->jsd, 0);
        axis1_value = JSGetAxisCoeff(&jc->jsd, 1);
        axis2_value = JSGetAxisCoeff(&jc->jsd, 2);	/* Rotate. */
        axis3_value = JSGetAxisCoeff(&jc->jsd, 3);	/* Throttle. */
        axis4_value = JSGetAxisCoeff(&jc->jsd, 4);
        axis5_value = JSGetAxisCoeff(&jc->jsd, 5);


	/* Hat. */
	w = lr->hat_da;
	pixmap = lr->hat_buf;
	gdk_window_get_size(pixmap, &width, &height);

        gdk_gc_set_foreground(gc, &lr->c_axis_bg);
        gdk_draw_rectangle(
            pixmap, gc,
            TRUE,
            0, 0,
            width, height
        );

        gdk_gc_set_foreground(gc, &lr->c_axis_fg);
        gdk_draw_rectangle(
            pixmap, gc,
            FALSE,
            0, 0,
            width, height
        );

        x = (width / 2);
        y = (height / 2);
        gdk_gc_set_foreground(gc, &lr->c_axis_grid);
	gdk_draw_line(
            pixmap, gc,
            x, 0,
            x, height - 1
	);
        gdk_draw_line( 
            pixmap, gc,
            0, y,
            width - 1, y
        );

	/* Draw hat. */
	if(total_axises >= 6)
	{
	    x = (axis4_value * (width / 4)) + (width / 2);
	    y = (gfloat)height - (
	        (axis5_value * (height / 4)) + (height / 2)
	    );
	}
	else if(total_axises == 5)
	{
            x = (axis3_value * (width / 4)) + (width / 2);
            y = (gfloat)height - (
                (axis4_value * (height / 4)) + (height / 2)
            );
	}
        else
        {
            x = (width / 2);
            y = height - (height / 2);
        }
        gdk_gc_set_foreground(gc, &lr->c_axis_hat);
	gdk_draw_arc(
	    pixmap, gc,
	    TRUE,
	    x - 10, y - 10,
	    20, 20,
	    0 * 64, 360 * 64
	);

        gdk_draw_pixmap(      
            w->window, 
            gc,  
            pixmap,
            0, 0,
            0, 0,
            width, height
        );


	/* First two axises. */
        w = lr->axis_da;
	pixmap = lr->axis_buf;
        gdk_window_get_size(pixmap, &width, &height);

        gdk_gc_set_foreground(gc, &lr->c_axis_bg);
        gdk_draw_rectangle(
            pixmap, gc,
            TRUE,
            0, 0,
            width, height
        );

        gdk_gc_set_foreground(gc, &lr->c_axis_fg);
        gdk_draw_rectangle(
            pixmap, gc,
            FALSE,
            0, 0,
            width, height
        );

        /* Null zone circle and grids. */
        x = (width / 2);
        y = (height / 2);
        gdk_gc_set_foreground(gc, &lr->c_axis_grid);
        axis_ptr1 = JCGetAxis(&jc->jsd, 0);
        axis_ptr2 = JCGetAxis(&jc->jsd, 1);
        if((axis_ptr1 != NULL) &&
           (axis_ptr2 != NULL)
        )
        {
            gint axis_len1, win_len1;
            gint axis_len2, win_len2;
  
            axis_len1 = MAX(axis_ptr1->max - axis_ptr1->min, 1);
            axis_len2 = MAX(axis_ptr2->max - axis_ptr2->min, 1);
            win_len1 = (gfloat)width * (
                (gfloat)axis_ptr1->nz / (gfloat)axis_len1
            );
            win_len2 = (gfloat)height * (
                (gfloat)axis_ptr2->nz / (gfloat)axis_len2
            );

            gdk_draw_arc(
                pixmap, gc,
                FALSE,
                x - (win_len1 / 2), y - (win_len2 / 2),
                win_len1, win_len2,
                0 * 64, 360 * 64
            );
        }
        gdk_draw_line(
            pixmap, gc,
            x, 0,
            x, height - 1
        );
        gdk_draw_line(
            pixmap, gc,
            0, y,
            width - 1, y
        );

	x = (axis0_value * (gfloat)width / 2) + ((gfloat)width / 2);
	y = (gfloat)height - (gfloat)(
	    (axis1_value * (gfloat)height / 2) + ((gfloat)height / 2)
	);
        gdk_gc_set_foreground(gc, &lr->c_axis_fg);
        gdk_draw_line(
            pixmap, gc,
            x - 5, y, x + 5, y
        );
        gdk_draw_line(
            pixmap, gc,
            x, y - 5, x, y + 5
        );

        gdk_draw_pixmap(
            w->window,
            gc,
            pixmap,
            0, 0,
            0, 0,
            width, height
        );


        /* Throttle. */
        w = lr->throttle_da;
        pixmap = lr->throttle_buf;
        gdk_window_get_size(pixmap, &width, &height);

        gdk_gc_set_foreground(gc, &lr->c_axis_bg);
        gdk_draw_rectangle(
            pixmap, gc,
            TRUE,
            0, 0,
            width, height
        );
        gdk_gc_set_foreground(gc, &lr->c_axis_fg);
        gdk_draw_rectangle(
            pixmap, gc,
            FALSE,
            0, 0,
            width, height
        );

	if((total_axises >= 6) || (total_axises == 4))
	{
            y = height - (
                (axis3_value * height / 2) + (height / 2)
	    );
	}
	else if((total_axises == 5) || (total_axises == 3))
	{
	    y = height - (
                (axis2_value * height / 2) + (height / 2)
            );
	}
	else
	{
	    y = height - 0;
	}
	gdk_gc_set_foreground(gc, &lr->c_throttle_fg);
        gdk_draw_rectangle(
            pixmap, gc,
            TRUE,
            0, y,
            width, height - y
        );

        gdk_draw_pixmap(
            w->window, gc,
            pixmap,
            0, 0,
            0, 0,
            width, height
        );


        /* Rotate. */
        w = lr->rotate_da;
        pixmap = lr->rotate_buf;
        gdk_window_get_size(pixmap, &width, &height);

        gdk_gc_set_foreground(gc, &lr->c_axis_bg);
        gdk_draw_rectangle(
            pixmap, gc,
            TRUE,
            0, 0,
            width, height
        );

        gdk_gc_set_foreground(gc, &lr->c_axis_fg);
        gdk_draw_rectangle(
            pixmap, gc,
            FALSE,
            0, 0,
            width, height
        );

        gdk_gc_set_foreground(gc, &lr->c_axis_grid);
	x = width / 2;
        axis_ptr1 = JCGetAxis(&jc->jsd, 2);
        if(axis_ptr1 != NULL)
        {
            gint axis_len1, win_len1;

            axis_len1 = MAX(axis_ptr1->max - axis_ptr1->min, 1);
            win_len1 = (gfloat)width * (
                (gfloat)axis_ptr1->nz / (gfloat)axis_len1
            );

            gdk_draw_line(
		pixmap, gc,
                x - (win_len1 / 2), 1,
		x - (win_len1 / 2), height - 1
	    );
            gdk_draw_line(
                pixmap, gc,
                x + (win_len1 / 2), 1,
                x + (win_len1 / 2), height - 1
            );
        }
        gdk_draw_line(
            pixmap, gc,
            x, 1,
            x, height - 1
        );

	if((total_axises >= 6) || (total_axises == 4))
	{
	    x = (axis2_value * width / 2) + (width / 2);
	}
	else
	{
	    x = width / 2;
	}
        gdk_gc_set_foreground(gc, &lr->c_axis_fg);
        gdk_draw_line(
            pixmap, gc,
            x, 1,
	    x, height - 1
        );

        gdk_draw_pixmap(
            w->window, gc,
            pixmap,
            0, 0,
            0, 0,
            width, height
        );
	gdk_flush();

	return;
}

/*
 *      Redraws the axises layout on the given jc 
 *      for the logical layout.  The given inputs are
 *      assumed valid and the layout is assumed to be logical.
 */
void JCDrawAxisesLogical(jc_struct *jc)
{
        gint i, x, width, height;
	gfloat axis_value;
        GtkWidget *w;
        GdkPixmap *pixmap;
	GdkGC *gc;
	layout_logical_struct *ll;
	layout_logical_axis_struct *ll_axis_ptr;
	js_axis_struct *axis_ptr;


        ll = &jc->layout_logical;

        /* Not initialized? */
        if(ll->toplevel == NULL)
            return;

        gc = jc->gc;
        if(gc == NULL)
            return;


	/* Go through each axis. */
	for(i = 0; i < ll->total_axises; i++)
	{
	    ll_axis_ptr = ll->axis[i];

	    if((ll_axis_ptr->position_buf == NULL) ||
               (ll_axis_ptr->position_da == NULL)
	    )
		continue;

            /* Get axis value coefficient. */
            axis_value = JSGetAxisCoeff(&jc->jsd, i);

            w = ll_axis_ptr->position_da;
            pixmap = ll_axis_ptr->position_buf;
            gdk_window_get_size(pixmap, &width, &height);

            gdk_gc_set_foreground(gc, &ll_axis_ptr->c_axis_bg);
            gdk_draw_rectangle(
                pixmap, gc,
                TRUE,
                0, 0,
                width, height
            );
            gdk_gc_set_foreground(gc, &ll_axis_ptr->c_axis_fg);
            gdk_draw_rectangle(
                pixmap, gc,
                FALSE,
                0, 0,
                width, height
            );

            x = width / 2;
            gdk_gc_set_foreground(gc, &ll_axis_ptr->c_axis_grid);
            gdk_draw_line(
                pixmap, gc,
                x, height - 8,
                x, height - 4
            );
            gdk_draw_line(
                pixmap, gc,
                1, height - 3,
		width - 1, height - 3
            );

	    axis_ptr = JCGetAxis(&jc->jsd, i);
            if(axis_ptr != NULL)
            {
                gint axis_len1, win_len1;

                axis_len1 = MAX(axis_ptr->max - axis_ptr->min, 1);
                win_len1 = (gfloat)width * (
                    (gfloat)axis_ptr->nz / (gfloat)axis_len1
                );

                gdk_draw_rectangle(
                    pixmap, gc,
                    TRUE,
		    x - (win_len1 / 2), height - 5,
                    win_len1, 2
                );

                gdk_draw_line(
                    pixmap, gc,
                    x - (win_len1 / 2), height - 7,
		    x - (win_len1 / 2), height - 3
                );
                gdk_draw_line(
                    pixmap, gc,
                    x + (win_len1 / 2), height - 7,
                    x + (win_len1 / 2), height - 3
                );
	    }

            x = (axis_value * (gfloat)width / 2) + ((gfloat)width / 2);
            gdk_gc_set_foreground(gc, &ll_axis_ptr->c_axis_fg);
	    if(1)
	    {
		GdkPoint p[3];

		/* Draw arrow poly. */
		if(x <= 0)
		{
		    x = 0;
                    p[0].x = x + 4;
                    p[0].y = height - 13;
                    p[1].x = x;
                    p[1].y = height - 10;
                    p[2].x = x + 4;
                    p[2].y = height - 7;
		}
		else if(x >= (width - 1))
		{
		    x = width - 1;
		    p[0].x = x - 4;
                    p[0].y = height - 13;
                    p[1].x = x;
                    p[1].y = height - 10;
                    p[2].x = x - 4;
                    p[2].y = height - 8;
		}
		else
		{
		    p[0].x = x - 3;
		    p[0].y = height - 11;
		    p[1].x = x + 3;
                    p[1].y = height - 11;
		    p[2].x = x;
                    p[2].y = height - 7;
		}

                gdk_draw_polygon(
		    pixmap, gc,
		    TRUE,
		    p, 3
		);
	    }

	    x = (width / 2) - (width / 4);
            gdk_draw_line(
                pixmap, gc,
                x, height - 7,
                x, height - 3
            );
            x = (width / 2) + (width / 4);
            gdk_draw_line(
                pixmap, gc,
                x, height - 7,
                x, height - 3
            );

            gdk_draw_pixmap(
                w->window, gc,
                pixmap,
                0, 0,
                0, 0,
                width, height
            );
	}
        gdk_flush();

	return;
}


/*
 *	Redraws the axises layout on te given jc depending
 *	on it's set layout style.
 */
void JCDrawAxises(jc_struct *jc)
{
	if(jc == NULL)
	    return;
	if(jc->toplevel == NULL)
	    return;

	/* Handle by currently selected axis layout page. */
	switch(jc->layout_state)
	{
	  case JC_LAYOUT_REPRESENTATIVE:
	    JCDrawAxisesRepresentative(jc);
	    break;

	  case JC_LAYOUT_LOGICAL:
	    JCDrawAxisesLogical(jc);
	    break;
	}

	return;
}

/*
 *	Redraw button labels.
 */
void JCDrawButtons(jc_struct *jc)
{
	gint i, total;
	buttons_list_struct *bl;
	js_button_struct **btn;
	GtkWidget *w;


	if(jc == NULL)
	    return;

	bl = &jc->buttons_list;
        if(bl == NULL)
            return;


	total = bl->total_buttons;
	if(total > jc->jsd.total_buttons)
	    total = jc->jsd.total_buttons;

	for(i = 0, btn = jc->jsd.button; i < total; i++, btn++)
	{
	    w = bl->button[i];
	    if(w == NULL)
		continue;

	    if(*btn == NULL)
		continue;


	    GTK_BUTTON(w)->in_button = (((*btn)->state) ? 1 : 0);
	    gtk_signal_emit_by_name(
		GTK_OBJECT(w),
		(((*btn)->state) ? "pressed" : "released")
	    );
	}


	return;
}


/*
 *	Redraw signal level.
 */
void JCDrawSignalLevel(jc_struct *jc)
{
	gint i, width, height;
	u_int16_t *history;
        GtkWidget *w;
	GdkPixmap *pixmap;
	GdkGC *gc;


        if(jc == NULL)
            return;

        gc = jc->gc;
	w = jc->signal_level_da;
	pixmap = jc->signal_level_pixmap;
	if((w == NULL) ||
           (pixmap == NULL)
	)
	    return;

        gdk_window_get_size(pixmap, &width, &height);

        gdk_gc_set_foreground(gc, &jc->c_signal_level_bg);
        gdk_draw_rectangle(
            pixmap, gc,
            TRUE,
            0, 0,
            width, height
        );

        gdk_gc_set_foreground(gc, &jc->c_signal_level_fg);
	history = jc->signal_history;
	for(i = 0; i < JC_SIGNAL_HISTORY_MAX; i++)
	{
	    gdk_draw_line(
		pixmap, gc,
		i,
		height - (
		    height * ((gfloat)history[i] /
			(gfloat)((u_int16_t)-1)
		    )
		) - 1,
		i,
		height - 1
	    );
	}

        gdk_draw_pixmap(
            w->window, gc,
            pixmap,
            0, 0,
            0, 0,
            width, height
        );

	return;
}

