#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/disk.h"

#include "jc.h"


gint JCCreate(jc_struct *jc, gint argc, gchar **argv);
void JCMap(jc_struct *jc);
void JCUnmap(jc_struct *jc);
void JCDestroy(jc_struct *jc);


/*
 *	Procedure to create the given jc.
 */
gint JCCreate(jc_struct *jc, gint argc, gchar **argv)
{
	gint i, status;
	gchar *arg, *strptr, *strptr2;


	if(jc == NULL)
	    return(-1);


	/* Reset joystick data structure values. */
	memset(&jc->jsd, 0x00, sizeof(js_data_struct));
	jc->jsd.fd = -1;

	/* Set default joystick calibration file path. */
	strptr = getenv("HOME");
	if(strptr == NULL)
	     strptr2 = JC_DEF_JS_CALIB;
	else
	     strptr2 = PrefixPaths(strptr, JC_DEF_JS_CALIB);
	jc->calib_file = StringCopyAlloc(strptr2);
	jc->tried_load_device_on_init = FALSE;

	/* Reset has changes marker. */
	jc->has_changes = FALSE;


        /* Create widgets. */
        status = JCCreateWidgets(jc, argc, argv);
        if(status)
            return(-1);


	/* Handle arguments. */
	for(i = 1; i < argc; i++)
	{
	    arg = argv[i];
	    if(arg == NULL)
		continue;

	    /* Specify alternate calibration file. */
	    if(strcasepfx(arg, "-f"))
	    {
		i++;
		if(i < argc)
		{
		    free(jc->calib_file);
		    jc->calib_file = StringCopyAlloc(argv[i]);
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			argv[i - 1]
		    );
		}
	    }
	    /* Specify startup device. */
	    else if(strcasepfx(arg, "-d"))
	    {
                i++;
                if(i < argc)
                {
		    GtkCombo *combo = (GtkCombo *)jc->js_device_combo;

		    if(combo != NULL)
			gtk_entry_set_text(
			    GTK_ENTRY(combo->entry),
			    argv[i]
		        );
		}
                else
                {
                    fprintf(
                        stderr,
                        "%s: Requires argument.\n",
                        argv[i - 1] 
                    );
                }
	    }
	}


	return(0);
}

/*
 *	Maps the given jc.
 */
void JCMap(jc_struct *jc) 
{
        if(jc == NULL)
            return;

	if(jc->toplevel != NULL)
	    gtk_widget_show(jc->toplevel);

	return;
}

/*
 *	Unmaps the given jc.
 */
void JCUnmap(jc_struct *jc)
{
        if(jc == NULL)
            return;

        if(jc->toplevel != NULL)
            gtk_widget_hide(jc->toplevel);

	return;
}

/*
 *	Procedure to deallocate all resources of the given jc.
 */
void JCDestroy(jc_struct *jc)  
{
	if(jc == NULL)
	    return;

	JCDoCloseJoystick(jc);

	JCDestroyWidgets(jc);

	free(jc->calib_file);
        jc->calib_file = NULL;


	return;
}
