#include <stdio.h>
#include <signal.h>

#include "../include/string.h"

#include "cdialog.h"
#include "pdialog.h"
#include "jc.h"


gint JCInit(gint argc, gchar **argv);
void JCShutdown(void);


jc_cursor_struct jc_cursor;
jc_struct js_calib;
GtkWidget *filebrowser;


/*
 *	Procedure to initialize all resources.
 */
gint JCInit(gint argc, gchar **argv)
{
	gint i, status;
	gchar *arg;


        signal(SIGINT, JCSignalHandler);
        signal(SIGTERM, JCSignalHandler);
        signal(SIGSEGV, JCSignalHandler);



	/* Parse arguments. */
	for(i = 0; i < argc; i++)
	{
	    arg = argv[i];
	    if(arg == NULL)
		continue;

	    /* Help. */
	    if(strcasepfx(arg, "--h") ||
               strcasepfx(arg, "-h") ||
               strcasepfx(arg, "-?")
	    )
	    {
		printf(PROG_USAGE_MESG);
		return(-4);
	    }
	    /* Version. */
	    else if(strcasepfx(arg, "--v") ||
                    strcasepfx(arg, "-v")
	    )
	    {
		printf(
		    "%s Version %s\n%s\n",
		    PROG_NAME, PROG_VERSION,
		    PROG_COPYRIGHT
		);
		return(-4);
	    }
	}


	/* Initialize toolkit. */
        gtk_init(&argc, &argv);


	/* Load cursors. */
	jc_cursor.processing = gdk_cursor_new(GDK_WATCH);

	/* Initialize cdialog. */
	CDialogInit();

	/* Initialize prompt dialog. */
	PDialogInit();

	/* Create joystick calibrator window and map it. */
	status = JCCreate(&js_calib, argc, argv);
	if(status)
	    return(-1);
	JCMap(&js_calib);


	/* Create file browser. */
	status = JCFBCreate(&filebrowser, argc, argv);
	if(status)
	    return(-1);

        gtk_timeout_add(
            16,			/* Every 16 milliseconds. */
            JCTimeoutCB,
            NULL
        );

	return(0);
}

/*
 *	Procedure to deallocate all resources.
 */
void JCShutdown(void)
{
	GdkCursor **cursor_ptr;


	/* Destroy everything we allocated. */

	/* File browser. */
        JCFBDestroy(&filebrowser);

	/* Joystick caliberator. */
        JCDestroy(&js_calib);

	/* Prompt dialog. */
	PDialogShutdown();

	/* Cdialog. */
	CDialogShutdown();

	/* Cursors. */
#define DO_DESTROY_CURSOR	\
{ \
 if((*cursor_ptr) != NULL) \
 { \
  gdk_cursor_destroy(*cursor_ptr); \
  (*cursor_ptr) = NULL; \
 } \
}
	cursor_ptr = &jc_cursor.processing;
	DO_DESTROY_CURSOR

#undef DO_DESTROY_CURSOR

	return;
}


gint main(gint argc, gchar **argv)
{
	gint status;


	status = JCInit(argc, argv);
	switch(status)
	{
	  case 0:
	    break;

	  case -4:
	    /* No shutdown of resources. */
	    return(0);
	    break;

	  default:
	    JCShutdown();
            return(-1);
	    break;
	}

        gtk_main();

	return(0);
}
